/**
  * File : Conv.cpp
  * Desc : 
  * Author : daizongh
  * Version : V1.0 
  * Date : 2007-05-30 
  */
#include "Conv.h"
#include "conv_tbl.h"
//#include <inttypes.h>

namespace d_tools
{
	U16CHAR gbk_to_utf16(U16CHAR c)
	{
		U16CHAR ret = 0;
		if (c >= GBK_START && c <= GBK_END)
		{
			ret = gbk2unicode[c - GBK_START];
		}
		if (ret == 0) return defUniChar;
		return ret;
	}

	U16CHAR utf16_to_gbk(U16CHAR c)
	{
		if( unicode2gbk[c - 1] != 0 ) return unicode2gbk[c - 1];
		return defGbkChar;
	}

	U16CHAR big5_to_utf16(U16CHAR c)
	{
		U16CHAR ret = 0;
		if( c >= BIG_START && c <= BIG_END )
		{
			ret = big2unicode[c - BIG_START];
		}
		if( ret == 0 ) return defUniChar;
		return ret;
	}

	U16CHAR utf16_to_big5(U16CHAR c)
	{
		if( unicode2big[c - 1] != 0 ) return unicode2big[c - 1];
		return defBigChar;
	}

	int gbk_to_utf16(U8CHAR * src, unsigned int len, U16CHAR * dst, unsigned int buf_len)
	{
		unsigned int odx = 0;
		for( unsigned int i = 0; i < len; i++ )
		{
			if( odx >= buf_len ) break;
			if( src[i] < 0x80 ) dst[odx] = src[i];
			else
			{
				dst[odx] = gbk_to_utf16((src[i] << 8) | src[i + 1]);
				i++;
			}
			odx++;
		}
		return odx;
	}

	int utf16_to_gbk(U16CHAR * src, unsigned int len, U8CHAR * dst, unsigned int buf_len)
	{
		U16CHAR val;
		unsigned int odx = 0;
		for( unsigned int i = 0; i < len; i++ )
		{
			val = utf16_to_gbk( src[i] );
			if( val < 0x80 ) 
			{
				if( odx >= buf_len ) break;
				dst[odx++] = val;
			}
			else
			{
				if( odx + 1 >= buf_len ) break;
				dst[odx++] = val >> 8;
				dst[odx++] = val & 0xFF;
			}
		}
		return odx;
	}

	int big5_to_utf16(U8CHAR * src, unsigned int len, U16CHAR * dst, unsigned int buf_len)
	{
		unsigned int odx = 0;
		for( unsigned int i = 0; i < len; i++ )
		{
			if( odx >= buf_len ) break;
			if( src[i] < 0x80 ) dst[odx++] = src[i];
			else
			{
				dst[odx++] = big5_to_utf16( (src[i] << 8) | src[i + 1] );
				i++;
			}
		}
		return odx;
	}

	int utf16_to_big5(U16CHAR * src, unsigned int len, U8CHAR * dst, unsigned int buf_len)
	{
		U16CHAR val;
		unsigned int odx = 0;
		for( unsigned int i = 0; i < len; i++ )
		{
			val = utf16_to_big5( src[i] );
			if( val < 0x80 )
			{
				if( odx >= buf_len ) break;
				dst[odx++] = val;
			}
			else
			{
				if( odx + 3 >= buf_len ) break;
				dst[odx++] = val >> 8;
				dst[odx++] = val & 0xFF;
			}
		}
		return odx;
	}


	/********* utf8 <=> utf16 **********/
#define UNI_SUR_HIGH_START      (unsigned int)0xD800
#define UNI_SUR_HIGH_END        (unsigned int)0xDBFF
#define UNI_SUR_LOW_START       (unsigned int)0xDC00
#define UNI_SUR_LOW_END         (unsigned int)0xDFFF

	static const int halfShift      = 10; /* used for shifting by 10 bits */
	static const unsigned int halfBase     = 0x0010000UL;
	static const unsigned int halfMask     = 0x3FFUL;
	static const char trailingBytesForUTF8[256] = {
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
		2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5
	};
	static const unsigned int offsetsFromUTF8[6] = {
		0x00000000UL, 0x00003080UL, 0x000E2080UL,
		0x03C82080UL, 0xFA082080UL, 0x82082080UL
	};
	static const unsigned char firstByteMark[7] = {
		0x00, 0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC
	};

	static bool isLegalUTF8(const U8CHAR *source, int length)
	{
		U8CHAR a;
		const U8CHAR * srcptr = source+length;
		switch (length) 
		{
			default: return false;
					 /* Everything else falls through when "true"... */
			case 4: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return false;
			case 3: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return false;
			case 2: if ((a = (*--srcptr)) > 0xBF) return false;
						switch (*source) {
							/* no fall-through in this inner switch */
							case 0xE0: if (a < 0xA0) return false; break;
							case 0xF0: if (a < 0x90) return false; break;
							case 0xF4: if (a > 0x8F) return false; break;
							default:  if (a < 0x80) return false;
						}
			case 1: if (*source >= 0x80 && *source < 0xC2) return false;
						if (*source > 0xF4) return false;
		}
		return true;
	}

	int utf8_to_utf16(U8CHAR * src, unsigned int src_len, U16CHAR * dest, unsigned int dest_len)
	{
		U8CHAR * srcEnd = src + src_len;
		U16CHAR * destStart = dest;
		U16CHAR * destEnd = dest + dest_len;
		while (src < srcEnd && dest < destEnd)
		{
			unsigned long ch = *src;
			U16CHAR extraBytesToRead = trailingBytesForUTF8[ch];
			if (src + extraBytesToRead >= srcEnd)
			{
				break;
			}
			if (isLegalUTF8(src, extraBytesToRead + 1))
			{
				ch = 0;
				switch (extraBytesToRead) 
				{
					case 3: ch += *src++; ch <<= 6;
					case 2: ch += *src++; ch <<= 6;
					case 1: ch += *src++; ch <<= 6;
					case 0: ch += *src++;
				}
				ch -= offsetsFromUTF8[extraBytesToRead];
			}
			else 
			{
				return -1;
			}
			if (ch < 0x10000)
			{
				*dest++ = ch;
			}
			else if (dest + 2 < destEnd)
			{
				*dest++ = 0xD800 | (((ch) >> 10) - 0x0040);
				*dest++ = 0xDC00 | ((ch) & 0x03FF);
			}
			else 
			{
				return -1; 
			}
		}
		return dest - destStart;
	}

	int utf16_to_utf8(U16CHAR * src, unsigned int srcLen, U8CHAR * dest, unsigned int destLen)
	{
		const U16CHAR * srcEnd = src + srcLen;
		U8CHAR * destStart = dest;
		U8CHAR * destEnd = dest + destLen;

		while (src < srcEnd){
			unsigned int ch;
			U16CHAR bytesToWrite = 0;
			const unsigned int byteMask = 0xBF;
			const unsigned int byteMark = 0x80;

			ch = *src++;
			if (ch >= UNI_SUR_HIGH_START && ch <= UNI_SUR_HIGH_END &&
					src < srcEnd) 
			{
				unsigned int ch2 = *src;
				if (ch2 >= UNI_SUR_LOW_START && ch2 <= UNI_SUR_LOW_END) 
				{
					ch = ((ch - UNI_SUR_HIGH_START) << halfShift)
						+ (ch2 - UNI_SUR_LOW_START) + halfBase;
					++src;
				}
				else
				{
					return -1;
				}
			} 
			else if (ch >= UNI_SUR_LOW_START && ch <= UNI_SUR_LOW_END) 
			{
				return -1;
			}
			/* Figure out how many bytes the result will require */
			if (ch < (unsigned int)0x80) {                 bytesToWrite = 1;
			} else if (ch < (unsigned int)0x800) {         bytesToWrite = 2;
			} else if (ch < (unsigned int)0x10000) {       bytesToWrite = 3;
			} else if (ch < (unsigned int)0x200000) {      bytesToWrite = 4;
			} else { return -1;
			}
			dest += bytesToWrite;
			if (dest < destEnd)
			{
				switch (bytesToWrite)
				{ /* note: everything falls through. */
					case 4:     *--dest = (ch | byteMark) & byteMask; ch >>= 6;
					case 3:     *--dest = (ch | byteMark) & byteMask; ch >>= 6;
					case 2:     *--dest = (ch | byteMark) & byteMask; ch >>= 6;
					case 1:     *--dest =  ch | firstByteMark[bytesToWrite];
				}
				dest += bytesToWrite;

			}
			else 
			{
				return -1;
			}
		}
		return dest - destStart;
	}
#undef UNI_SUR_HIGH_START
#undef UNI_SUR_HIGH_END
#undef UNI_SUR_LOW_START
#undef UNI_SUR_LOW_END
}
